--------------------------------------------------------------------------------
--
--   FileName:         bit_serial_mult.vhd
--   Dependencies:     dff.vhd, full_adder.vhd, my_and2.vhd
--
--   HDL CODE IS PROVIDED "AS IS."  DIGI-KEY EXPRESSLY DISCLAIMS ANY
--   WARRANTY OF ANY KIND, WHETHER EXPRESS OR IMPLIED, INCLUDING BUT NOT
--   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
--   PARTICULAR PURPOSE, OR NON-INFRINGEMENT. IN NO EVENT SHALL DIGI-KEY
--   BE LIABLE FOR ANY INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL
--   DAMAGES, LOST PROFITS OR LOST DATA, HARM TO YOUR EQUIPMENT, COST OF
--   PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
--   BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF),
--   ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER SIMILAR COSTS.
--
--   Version History
--   Version 1.0 12/14/16 -TR
--
--------------------------------------------------------------------------------


--Bit serial N-bit x N-bit multiplier

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;


entity bit_serial_mult is
  generic (data_width : natural := 8);
  port (
    multiplicand  : in std_logic_vector(data_width-1 downto 0);
    multiplier    : in std_logic;
    rst_n         : in std_logic;
    clk           : in std_logic;
    product       : out std_logic
  );
end entity bit_serial_mult;

architecture arch of bit_serial_mult is
  --Internal signals and components
  signal and_outs       : std_logic_vector(data_width-1 downto 0);
  signal adder_outs     : std_logic_vector(data_width-1 downto 0);
  signal adder_reg_outs : std_logic_vector(data_width-1 downto 0);
  signal carry_outs     : std_logic_vector(data_width-1 downto 0);
  signal carry_reg_outs : std_logic_vector(data_width-1 downto 0);

  signal product_adj : std_logic;
  signal s_multiplicand, multiplicand_complement : std_logic_vector(data_width-1 downto 0);
  signal intermediate_product_complement, complement : std_logic;

  component dfflop is
  	port (
  		clk	    :	in std_logic;
  		rst_n	  :	in std_logic;
  		D		    :	in std_logic;
  		Q		    :	out std_logic
  	);
  end component dfflop;

  component my_and2 is
    port(
      A : in std_logic;
      B : in std_logic;
      Z : out std_logic
    );
  end component my_and2;

  component full_adder is
    port (
      A     : in std_logic;
      B     : in std_logic;
      Cin   : in std_logic;
      S     : out std_logic;
      Cout  : out std_logic
    );
  end component full_adder;

begin

  --Make negative multiplicand positive
  multiplicand_complement <= std_logic_vector((not unsigned(multiplicand)) + 1);

  complement_product : process(adder_outs, clk, rst_n)
  begin
    if rst_n = '0' then
      complement <= '0';
    elsif rising_edge(clk) then
      if adder_outs(0) = '1' then
        complement <= '1';
      end if;
    end if;
  end process;

  --If multiplicand is negative, make it positive
  s_multiplicand <= multiplicand_complement when multiplicand(multiplicand'high) = '1' else multiplicand;

  --Create a 2's complement version of the product
  intermediate_product_complement <= not adder_outs(0) when complement = '1' else adder_outs(0);

  --Select either the 2's complemented product or the normal product based on the sign of the multiplicand
  product_adj <= intermediate_product_complement when multiplicand(multiplicand'high) = '1' else adder_outs(0);

  main_generate : for i in 0 to data_width-1 generate
    --Input AND gates
    and_gate : my_and2
    port map (
      A => s_multiplicand(i),
      B => multiplier,
      Z => and_outs(i)
    );

    general_adders : if i < data_width-1 generate
      full_adders : full_adder
      port map (
        A => and_outs(i),
        B => adder_reg_outs(i+1),
        Cin => carry_reg_outs(i),
        S => adder_outs(i),
        Cout => carry_outs(i)
      );
    end generate general_adders;

    --Special case for the MSB
    msb_adder : if i = data_width-1 generate
      msb_full_adder : full_adder
      port map (
        A => and_outs(i),
        B => '0',
        Cin => carry_reg_outs(i),
        S => adder_outs(i),
        Cout => carry_outs(i)
      );
    end generate msb_adder;

    --Carry flip flops
    carry_ffs : dfflop
    port map (
      clk => clk,
      rst_n => rst_n,
      D => carry_outs(i),
      Q => carry_reg_outs(i)
    );

    --Adder output flip flops
    adder_ffs : dfflop
    port map (
      clk => clk,
      rst_n => rst_n,
      D => adder_outs(i),
      Q => adder_reg_outs(i)
    );

  end generate main_generate;

	 --Output flip flop
	 output_ff : dfflop
	 port map (
		clk => clk,
		rst_n => rst_n,
		D => product_adj,
		Q => product
	 );

end architecture;
