--------------------------------------------------------------------------------
--
--   FileName:         serial_multiply_add.vhd
--   Dependencies:     bit_serial_mult.vhd, bit_serial_adder.vhd, dff.vhd
--
--   HDL CODE IS PROVIDED "AS IS."  DIGI-KEY EXPRESSLY DISCLAIMS ANY
--   WARRANTY OF ANY KIND, WHETHER EXPRESS OR IMPLIED, INCLUDING BUT NOT
--   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
--   PARTICULAR PURPOSE, OR NON-INFRINGEMENT. IN NO EVENT SHALL DIGI-KEY
--   BE LIABLE FOR ANY INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL
--   DAMAGES, LOST PROFITS OR LOST DATA, HARM TO YOUR EQUIPMENT, COST OF
--   PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
--   BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF),
--   ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER SIMILAR COSTS.
--
--   Version History
--   Version 1.1 11/03/20 - Fixed multiplier reset bug -TR
--   Version 1.0 12/14/16 -TR
--
--------------------------------------------------------------------------------

--Serial multiply-add block
--First bit(LSB) of the result is ready on the second clock cycle after starting the calculation
--Multiplier FFs are cleared after each operation is complete which is done when the enable signal drops low

library IEEE;
use IEEE.std_logic_1164.all;

entity serial_multiply_add is
  generic (
    data_width : natural := 256
  );
  port (
    clk     : in std_logic;
    enable  : in std_logic;
    rst_n   : in std_logic;
    X       : in std_logic_vector(data_width-1 downto 0);
    A       : in std_logic;
    B       : in std_logic;
    result  : out std_logic
  );
end entity;

architecture arch of serial_multiply_add is

  component bit_serial_mult is
    generic (data_width : natural := 8);
    port (
      multiplicand  : in std_logic_vector(data_width-1 downto 0);
      multiplier    : in std_logic;
      rst_n         : in std_logic;
      clk           : in std_logic;
      product       : out std_logic
    );
  end component;

  component bit_serial_adder is
    port (
      clk   : in std_logic;
      rst_n : in std_logic;
      a     : in std_logic;
      b     : in std_logic;
      s     : out std_logic
    );
  end component;

  component  dfflop is
  	port (
  		clk	:	in std_logic;
  		rst_n	:	in std_logic;
  		D		:	in std_logic;
  		Q		:	out std_logic
  	);
  end component;

  signal mult_result, ff_out : std_logic;
  signal internal_clk : std_logic;
  signal mult_rst_n : std_logic;

begin

  internal_clk <= clk and enable;
  mult_rst_n <= rst_n or enable;

  mult : bit_serial_mult
  generic map ( data_width )
  port map (
    multiplicand => X,
    multiplier => A,
    rst_n => mult_rst_n,
    clk => internal_clk,
    product => mult_result
  );

  input_ff : dfflop
  port map (
    clk => internal_clk,
    rst_n => rst_n,
    D => B,
    Q => ff_out
  );

  adder : bit_serial_adder
  port map (
    clk => internal_clk,
    rst_n => rst_n,
    a => mult_result,
    b => ff_out,
    s => result
  );

end architecture;
