--------------------------------------------------------------------------------
--
--   FileName:         serial_polynomial.vhd
--   Dependencies:     serial_multiply_add.vhd, dff.vhd, shift_register.vhd
--
--   HDL CODE IS PROVIDED "AS IS."  DIGI-KEY EXPRESSLY DISCLAIMS ANY
--   WARRANTY OF ANY KIND, WHETHER EXPRESS OR IMPLIED, INCLUDING BUT NOT
--   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
--   PARTICULAR PURPOSE, OR NON-INFRINGEMENT. IN NO EVENT SHALL DIGI-KEY
--   BE LIABLE FOR ANY INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL
--   DAMAGES, LOST PROFITS OR LOST DATA, HARM TO YOUR EQUIPMENT, COST OF
--   PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
--   BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF),
--   ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER SIMILAR COSTS.
--
--   Version History
--   Version 1.1 11/03/20 - Simplified coeff delay chain, modified reset timings -TR
--   Version 1.0 12/14/16 -TR
--
--------------------------------------------------------------------------------

--Serial polynomial top level

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity serial_polynomial is
  generic (
    data_width : natural := 32;
    poly_order : natural := 2
  );
  port (
    clk       : in std_logic;
    rst_n     : in std_logic;
    --Selects between serial and parallel input and output modes
    s_p       : in std_logic;   --Set high to use serial input/output
    --Begins operations, either serial load if s_p = 1 or starts calculation if s_p = 0 and data is available on the X port
    start     : in std_logic;
    --Polynomial inputs
    --Used to load the register serially if desired
    shift_X   : in std_logic;
    X         : in std_logic_vector(data_width-1 downto 0);
    --Polynomial coefficients
    coeffs    : in std_logic_vector(poly_order downto 0);
    --Done is asserted when the calculation is complete for one clock cycle
    done      : out std_logic;
    --Valid goes high once the first bit (LSB) of the result is shifted out and remains high until the operation is over and done is asserted
    valid     : out std_logic;
    result    : out std_logic_vector((poly_order + 1)*data_width-1 downto 0);
    --Serial output of the result
    shift_out : out std_logic
  );
end entity;

architecture arch of serial_polynomial is

  --Component declarations
  component serial_multiply_add is
    generic (
      data_width : natural := 32
    );
    port (
      clk     : in std_logic;
      enable  : in std_logic;
      rst_n   : in std_logic;
      X       : in std_logic_vector(data_width-1 downto 0);
      A       : in std_logic;
      B       : in std_logic;
      result  : out std_logic
    );
  end component;

  component shift_register is
    generic (data_width : natural := 8);
    port (
      rst_n       : in std_logic;
      clk         : in std_logic;
      s_p         : in std_logic;   --Selects between serial and parallel operation, serial input when high, parallel input when low
      dir         : in std_logic;   --Selects either a left or right shift, high for right, low for left
      reg_en_n    : in std_logic;   --Active low register enable, value in the register is held while this signal is high
      shift_in    : in std_logic;
      D           : in std_logic_vector(data_width-1 downto 0); --Parallel input
      Q           : out std_logic_vector(data_width-1 downto 0);
      serial_out  : out std_logic
    );
  end component;

  component dfflop is
  	port (
  		clk	:	in std_logic;
  		rst_n	:	in std_logic;
  		D		:	in std_logic;
  		Q		:	out std_logic
  	);
  end component;

  --End of component declarations

  --Signals for the X input register
  signal X_reg : std_logic_vector(data_width-1 downto 0);
  signal X_clr : std_logic;
  -- signal X_rst_n : std_logic;

  --Constants for determining completion status
  constant completion_cycles : integer := (2*poly_order) + (poly_order + 1)*data_width;
  constant lsb_cycles : integer := 2*poly_order;

  --State machine and control signals
  type state_type is (INIT, SERIAL_LOAD, CALC, COMPLETE);
  signal current_state : state_type;

  signal enable : std_logic_vector(poly_order downto 1);
  signal load_count : integer range 0 to data_width;
  signal calc_count : integer range 0 to completion_cycles;
  signal inc_load_count, inc_calc_count : std_logic;
  signal input_shift_reg_en_n : std_logic;
  signal output_shift_reg_en_n : std_logic;
  signal oe_flag : std_logic;

  --Intermediate signals for coefficients and results of each stage
  type coeff_array is array(0 to poly_order, 0 to 2*poly_order-2) of std_logic;
  signal intermediate_coeffs : coeff_array;

  signal intermediate_results : std_logic_vector(poly_order downto 1);

  signal not_clk : std_logic;

begin

  --Active low reset to clear the X input register after a calculation is complete
  -- X_rst_n <= not (not rst_n or X_clr);

  --Inverted clk to be used on coefficient registers
  not_clk <= not clk;

  valid <= not output_shift_reg_en_n;

  next_state_logic : process(clk, rst_n, current_state)
  begin
    if rst_n = '0' then
      current_state <= INIT;
      enable <= (others => '0');
      done <= '0';
      input_shift_reg_en_n <= '0';
      output_shift_reg_en_n <= '1';
      X_clr <= '0';
  		load_count <= 0;
  		calc_count <= 0;
      oe_flag <= '0';
      inc_load_count <= '0';
      inc_calc_count <= '0';

    elsif rising_edge(clk) then

      --Increment counters
      if inc_load_count = '1' then
        load_count <= load_count + 1;
      end if;

      if inc_calc_count = '1' then
        calc_count <= calc_count + 1;
      end if;

      case current_state is
        when INIT =>
          --Set default values
          inc_load_count <= '0';
          inc_calc_count <= '0';
          load_count <= 1;
          calc_count <= 1;  --Start at 1 to get correct timing
          oe_flag <= '0';
          enable <= (others => '0');

          done <= '0';
          X_clr <= '0';

          --Allow the input register to be loaded
          input_shift_reg_en_n <= '0';
          --Shifting not allowed on output
          output_shift_reg_en_n <= '1';

          if s_p = '1' and start = '1' then
            current_state <= SERIAL_LOAD;
            inc_load_count <= '1';

          elsif s_p = '0' and start = '1' then
            current_state <= CALC;
            enable(1) <= '1';
            inc_calc_count <= '1';

            --Prevent the X register from changing during operation
            input_shift_reg_en_n <= '1';

          else
            current_state <= INIT;

          end if;

        when SERIAL_LOAD =>

          --Enable shifting on the input register
          input_shift_reg_en_n <= '0';

          --Shifting not allowed on output register
          output_shift_reg_en_n <= '1';

          --Offset by one to get correct timing
          if load_count >= data_width - 1 then
            current_state <= CALC;
            enable(1) <= '1';
            inc_load_count <= '0';
            inc_calc_count <= '1';

            --Prevent the X register from changing during operation
            input_shift_reg_en_n <= '1';

          else
            current_state <= SERIAL_LOAD;

          end if;

        when CALC =>

          --Prevent the X register from changing during operation
          input_shift_reg_en_n <= '1';

          --Disable the serial loading counter
          inc_load_count <= '0';

          --If the result's LSB if ready, allow shifting to the output register
          --De-assert the done bit since the result will now change
          -- if oe_flag = '1' then
          --   output_shift_reg_en_n <= '0';
          --   done <= '0';
          -- end if;

          --Always enable first stage during calculation
          enable(1) <= '1';

          --Enable the next stage 2 clock cycles after the previous one
          if calc_count mod 2 = 0 then
            enable <= enable(poly_order-1 downto 1) & '1';
          end if;

          --Allow shifting of the output register
          --Offset by 2 to line it up with the correct clock edge
          -- if calc_count = lsb_cycles - 2 then
          --   oe_flag <= '1';

          if calc_count = lsb_cycles - 1 then
            output_shift_reg_en_n <= '0';
            done <= '0';

          --End calculation, offsetting by 1 to get the correct timing
          elsif calc_count >= completion_cycles - 1 then
            current_state <= COMPLETE;
            inc_calc_count <= '0';

            --Disable further shifting on output
            output_shift_reg_en_n <= '1';
            --Assert done bit
            done <= '1';

          else
            current_state <= CALC;

          end if;

        when COMPLETE =>

          --Assert done bit
          done <= '1';
          -- X_clr <= '1';

          --Disable further shifting on output
          output_shift_reg_en_n <= '1';

          --Disable counter
          inc_calc_count <= '0';
          current_state <= INIT;

        when others =>
          current_state <= INIT;

      end case;
    end if;
  end process;

  --Shift register for input X
  X_shift : shift_register
  generic map (data_width)
  port map (
    rst_n => rst_n,
    clk => clk,
    s_p => s_p,
    dir => '1',
    reg_en_n => input_shift_reg_en_n,
    shift_in => shift_X,
    D => X,
    Q => X_reg,
    serial_out => open
  );

  --Flip flops to delay coefficient inputs so they can all start at the same time
  coeff_ff_gen : for l in 0 to poly_order generate
    --Assign the direct coefficient inputs into the 0th column in the array
    intermediate_coeffs(l, 0) <= coeffs(l);

    ff_gen : for m in 1 to (2*poly_order - 2) generate
      coeff_ff : dfflop
      port map (
        clk => not_clk,
        rst_n => rst_n,
        D => intermediate_coeffs(l, m-1),
        Q => intermediate_coeffs(l, m)
      );
    end generate ff_gen;
  end generate coeff_ff_gen;

  --Generate multiply-add blocks for each polynomial stage
  gen_main : for k in 1 to poly_order generate
    first : if k = 1 generate
      first_stage : serial_multiply_add
      generic map (data_width)
      port map (
        clk => clk,
        enable => enable(1),
        rst_n => rst_n,
        X => X_reg,
        A => coeffs(0),
        B => coeffs(1),
        result => intermediate_results(1)
      );
    end generate first;

    general : if k > 1 generate
      general_stage : serial_multiply_add
      generic map (data_width)
      port map (
        clk => clk,
        enable => enable(k),
        rst_n => rst_n,
        X => X_reg,
        A => intermediate_results(k-1),
        B => intermediate_coeffs(k, 2*k-2),
        result => intermediate_results(k)
      );
    end generate general;
  end generate gen_main;

  --Assign the final result to the output, which is the MSB of the result vector
  shift_out <= intermediate_results(poly_order);

  --Output shift register
  out_shift : shift_register
  generic map ((poly_order + 1)*data_width)
  port map (
    rst_n => rst_n,
    clk => clk,
    s_p => '1',
    dir => '1',
    reg_en_n => output_shift_reg_en_n,
    shift_in => intermediate_results(poly_order),
    D => (others => '0'),
    Q => result,
    serial_out => open
  );

end architecture;
