--------------------------------------------------------------------------------
--
--   FileName:         tb_serial_polynomial.vhd
--   Dependencies:     serial_polynomial.vhd
--
--   HDL CODE IS PROVIDED "AS IS."  DIGI-KEY EXPRESSLY DISCLAIMS ANY
--   WARRANTY OF ANY KIND, WHETHER EXPRESS OR IMPLIED, INCLUDING BUT NOT
--   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
--   PARTICULAR PURPOSE, OR NON-INFRINGEMENT. IN NO EVENT SHALL DIGI-KEY
--   BE LIABLE FOR ANY INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL
--   DAMAGES, LOST PROFITS OR LOST DATA, HARM TO YOUR EQUIPMENT, COST OF
--   PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
--   BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF),
--   ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER SIMILAR COSTS.
--
--   Version History
--   Version 1.1 11/03/20 - Modified initialization timings - TR
--   Version 1.0 12/14/16 -TR
--
--------------------------------------------------------------------------------


--Exhaustive test bench for serial polynomial unit

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity tb_serial_polynomial_ex is
end entity;

architecture tb of tb_serial_polynomial_ex is

  component serial_polynomial is
    generic (
      data_width : natural := 32;
      poly_order : natural := 2
    );
    port (
      clk       : in std_logic;
      rst_n     : in std_logic;
      --Selects between serial and parallel input and output modes
      s_p       : in std_logic;   --Set high to use serial input/output
      --Begins operations, either serial load if s_p = 1 or starts calculation if s_p = 0 and data is available on the X port
      start     : in std_logic;
      --Polynomial input
      shift_X   : in std_logic;   --Used to load the register serially if desired
      X         : in std_logic_vector(data_width-1 downto 0);
      --Polynomial coefficients
      coeffs    : in std_logic_vector(poly_order downto 0);
      --Done is asserted when the calculation is complete for one clock cycle
      done      : out std_logic;
      --Valid is asserted while the shift output is valid
      valid     : out std_logic;
      result    : out std_logic_vector((poly_order + 1)*data_width-1 downto 0);
      shift_out : out std_logic
    );
  end component;

  constant W : natural := 8;
  constant order : natural := 3;

  signal clk : std_logic := '0';
  signal rst_n, s_p, start, shift_X, done, valid, shift_out : std_logic;
  signal coeffs : std_logic_vector(order downto 0);
  signal X : std_logic_vector(W-1 downto 0);
  signal result : std_logic_vector((order+1)*W-1 downto 0);

  signal expected_result: std_logic_vector((order+1)*W-1 downto 0);
  -- signal debug_flag : std_logic;

  type coeff_array is array(0 to order) of std_logic_vector(W-1 downto 0);
  signal coefficients : coeff_array;

begin

  UUT : serial_polynomial
  generic map (W, order)
  port map (clk, rst_n, s_p, start, shift_X, X, coeffs, done, valid, result, shift_out);

  clock : process
  begin
    clk <= not clk;
    wait for 20 ps;
  end process;

  inputs : process
    variable int_X, a, b, c, d : integer := 0;
    variable stage_index : integer := 0;
    variable temp_expected : integer := 0;
  begin

    --Set poly input and coefficients
    --Add terms as necessary for different ordered polynomials

    --Data Set 1

    int_X := -2;
    a := 2;
    b := 1;
    c := 5;
    d := 0;

    temp_expected := a*(int_X**order) + b*(int_x**(order-1)) + c*(int_X**(order-2)) + d*(int_X**(order-3));
    expected_result <= std_logic_vector(to_signed(temp_expected, expected_result'length));

    X <= std_logic_vector(to_signed(int_X, W));

    coefficients(0)   <= std_logic_vector(to_signed(a, W));
    coefficients(1)   <= std_logic_vector(to_signed(b, W));
    coefficients(2)   <= std_logic_vector(to_signed(c, W));
    coefficients(3)   <= std_logic_vector(to_signed(d, W));


    rst_n <= '0';
    start <= '0';
    shift_X <= '0';
    coeffs <= (others => '0');
    s_p <= '0';
    wait until falling_edge(clk);
    rst_n <= '1';

    wait until falling_edge(clk);
    start <= '1';
    shift_X <= '0';

    --Loop through the coefficients, changing them on each falling edge
    for l in 0 to W-1 loop
      for i in 0 to order loop
        coeffs(i) <= coefficients(i)(l);
      end loop;
      wait until falling_edge(clk);
    end loop;

    wait until rising_edge(done);
    wait until falling_edge(clk);

    start <= '0';

    --Data Set 2
    int_X := 8;
    a := 3;
    b := -4;
    c := 12;
    d := -16;

    temp_expected := a*(int_X**order) + b*(int_x**(order-1)) + c*(int_X**(order-2)) + d*(int_X**(order-3));
    expected_result <= std_logic_vector(to_signed(temp_expected, expected_result'length));

    X <= std_logic_vector(to_signed(int_X, W));

    coefficients(0)   <= std_logic_vector(to_signed(a, W));
    coefficients(1)   <= std_logic_vector(to_signed(b, W));
    coefficients(2)   <= std_logic_vector(to_signed(c, W));
    coefficients(3)   <= std_logic_vector(to_signed(d, W));

    rst_n <= '0';
    wait until falling_edge(clk);
    rst_n <= '1';
    s_p <= '0';

    wait until falling_edge(clk);
    start <= '1';
    shift_X <= '0';

    for l in 0 to W-1 loop
      for i in 0 to order loop
        coeffs(i) <= coefficients(i)(l);
      end loop;
      wait until falling_edge(clk);
    end loop;

    wait until rising_edge(done);
    wait until falling_edge(clk);

    start <= '0';


    --Data Set 3
    int_X := -5;
    a := -3;
    b := -31;
    c := 19;
    d := 45;

    temp_expected := a*(int_X**order) + b*(int_x**(order-1)) + c*(int_X**(order-2)) + d*(int_X**(order-3));
    expected_result <= std_logic_vector(to_signed(temp_expected, expected_result'length));

    X <= std_logic_vector(to_signed(int_X, W));

    coefficients(0)   <= std_logic_vector(to_signed(a, W));
    coefficients(1)   <= std_logic_vector(to_signed(b, W));
    coefficients(2)   <= std_logic_vector(to_signed(c, W));
    coefficients(3)   <= std_logic_vector(to_signed(d, W));

    rst_n <= '0';
    wait until falling_edge(clk);
    rst_n <= '1';
    s_p <= '0';

    wait until falling_edge(clk);
    start <= '1';
    shift_X <= '0';

    for l in 0 to W-1 loop
      for i in 0 to order loop
        coeffs(i) <= coefficients(i)(l);
      end loop;
      wait until falling_edge(clk);
    end loop;

    wait until rising_edge(done);
    wait until falling_edge(clk);

    start <= '0';

    --Data Set 4
    --Uses the serial mode to load X

    int_X := -12;
    a := 4;
    b := -15;
    c := -20;
    d := 9;

    temp_expected := a*(int_X**order) + b*(int_x**(order-1)) + c*(int_X**(order-2)) + d*(int_X**(order-3));
    expected_result <= std_logic_vector(to_signed(temp_expected, expected_result'length));

    X <= std_logic_vector(to_signed(int_X, W));
    coeffs <= (others => '0');

    coefficients(0)   <= std_logic_vector(to_signed(a, W));
    coefficients(1)   <= std_logic_vector(to_signed(b, W));
    coefficients(2)   <= std_logic_vector(to_signed(c, W));
    coefficients(3)   <= std_logic_vector(to_signed(d, W));

    rst_n <= '0';
    s_p <= '1';
    wait until falling_edge(clk);
    rst_n <= '1';

    wait until falling_edge(clk);
    start <= '1';

    coeffs(0) <= coefficients(0)(0);
    coeffs(1) <= coefficients(1)(0);
    coeffs(2) <= coefficients(2)(0);
    coeffs(3) <= coefficients(3)(0);

    for l in 0 to W-1 loop
      shift_X <= X(l);
      wait until falling_edge(clk);
    end loop;

    for l in 1 to W-1 loop
      for i in 0 to order loop
        coeffs(i) <= coefficients(i)(l);
      end loop;
      wait until falling_edge(clk);
    end loop;

    start <= '0';

    wait;


  end process;

end architecture;
